(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset2
imports "~~/src/HOL/Library/Multiset"
begin


lemma in_multiset_of: assumes "x \<in># multiset_of xs"
  shows "\<exists> bef aft. xs = bef @ x # aft"
proof -
  from assms[unfolded in_multiset_in_set set_conv_nth]
  obtain i where i: "i < length xs" and x: "xs ! i = x" by auto
  from id_take_nth_drop[OF i, unfolded x]
  show ?thesis by auto
qed

lemma in_multiset_of_eq: assumes "M + {#x#} = multiset_of ys"
  shows "\<exists> bef aft. ys = bef @ x # aft \<and> M = multiset_of (bef @ aft)"
proof -
  from assms[symmetric] have "x \<in># multiset_of ys" by auto
  from in_multiset_of[OF this] obtain bef aft where ys: "ys = bef @ x # aft"
    by auto
  have "M + {#x#} = multiset_of (bef @ aft) + {#x#}"
    unfolding assms ys multiset_eq_iff by auto
  hence "M = multiset_of (bef @ aft)" by auto
  from ys this show ?thesis by auto
qed

lemma multiset_of_map: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (map f xs) = multiset_of (map f ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

lemma multiset_of_concat: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (concat xs) = multiset_of (concat ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

definition list_of :: "'a multiset \<Rightarrow> 'a list"
  where "list_of M \<equiv> SOME xs. multiset_of xs = M"

lemma multiset_of_list_of[simp]: "multiset_of (list_of M) = M"
  by (unfold list_of_def, rule someI_ex, 
    insert surj_multiset_of[unfolded surj_def, THEN spec[of _ M]], 
    auto)

lemma list_of_singleton[simp]: "list_of {#x#} = [x]"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#x#}"
  then obtain y ys where xs: "xs = y # ys" by (cases xs, auto)
  show "xs = [x]" 
  proof (cases ys)
    case Nil
    with id xs show ?thesis by auto
  next
    case (Cons z zs)
    from id[unfolded xs Cons]
    have "multiset_of zs + {#z#} + {#y#} = {#x#}" by auto
    from this[unfolded union_is_single] show ?thesis by auto
  qed
qed

lemma list_of_empty[simp]: "list_of {#} = []"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#}"
  thus "xs = []" by (cases xs, auto)
qed
    
definition map :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a multiset \<Rightarrow> 'b multiset"
  where "map f M \<equiv> multiset_of (List.map f (list_of M))"

lemma map_union[simp]: "map f (M + N) = map f M + map f N"
proof -
  let ?xs = list_of
  have "multiset_of (List.map f (?xs (M + N)))
    = multiset_of (List.map f (?xs M @ ?xs N))"
    by (rule multiset_of_map,  auto)
  also have "... = multiset_of (List.map f (?xs M)) +
    multiset_of (List.map f (?xs N))"
    by auto
  finally show ?thesis 
    unfolding map_def .
qed

lemma map_empty[simp]: "map f {#} = {#}" 
  unfolding map_def
  by simp

lemma map_singleton[simp]: "map f {#x#} = {#f x#}" 
  unfolding map_def
  by simp

lemma map_list_of_multiset_of: "multiset_of (List.map f (list_of (multiset_of xs))) = multiset_of (List.map f xs)"
  by (rule multiset_of_map, simp)


definition concat :: "'a multiset multiset \<Rightarrow> 'a multiset"
  where "concat M \<equiv> multiset_of (List.concat (List.map list_of (list_of M)))"

lemma concat_union[simp]: "concat (M + N) = concat M + concat N"
proof -
  let ?xs = list_of
  have "multiset_of (List.concat (List.map ?xs (?xs (M + N))))
       = multiset_of (List.concat (List.map ?xs (?xs M @ ?xs N)))"
    by (rule multiset_of_concat, rule multiset_of_map, simp)
  also have "... = multiset_of (List.concat (List.map ?xs (?xs M))) 
    + multiset_of (List.concat (List.map ?xs (?xs N)))"
    by auto
  finally show ?thesis unfolding concat_def .
qed

lemma concat_empty[simp]: "concat {#} = {#}" 
  unfolding concat_def
  by simp

lemma concat_singleton[simp]: "concat {#M#} = M" 
  unfolding concat_def
  by simp

lemma map_map: "map f (map g M) = map (f o g) M"
  unfolding map_def map_list_of_multiset_of map_map 
  by simp

lemma map_concat: "map f (concat Ms) = concat (map (map f) Ms)"
proof -
  obtain xs where xs: "list_of Ms = xs" by auto
  show ?thesis
    unfolding concat_def map_def[of f]
    unfolding map_list_of_multiset_of
    unfolding map_concat
    unfolding List.map_map o_def
  proof (rule multiset_of_concat, unfold map_def[symmetric], unfold map_map o_def, unfold map_def[of _ Ms] xs)
    oops (* unsolvable, problem: multiset_of_concat is too restrictive *)


hide_fact(open) map_map
hide_const(open) map
hide_const(open) concat

end
