(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Impl
imports
  Term
  Monadic_Check
begin

text {*Resolve problems with recurisve dependencies in generated code.*}
lemma [code]:
  shows "HOL.eq (Var x) (Var y) \<longleftrightarrow> HOL.eq x y"
  and "HOL.eq (Fun f ts) (Fun g us) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ts us"
  by (simp_all add: list_all2_eq [symmetric])

subsection {* Useful abstractions *}

text {* Given that we perform the same operations on terms in order to get
a list of the variables and a list of the functions, we define functions
that run through the term and perform these actions. *}

fun contains_var_term :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "contains_var_term x (Var y) = (x = y)"
| "contains_var_term x (Fun _ ts) = list_ex (contains_var_term x) ts"

lemma contains_var_term_sound[simp]:
  "contains_var_term x t \<longleftrightarrow> x \<in> vars_term t"
  by (induct t) (auto simp: Bex_set_list_ex[symmetric])

(*use efficient implementation for code-generation*)
lemma [code_inline]: "x \<in> vars_term t = contains_var_term x t" by simp

fun vars_term_list' :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list' (Var x) = [x]"
| "vars_term_list' (Fun _ ts) = concat (map vars_term_list' ts)"

lemma vars_term_list'_sound[simp]: "set (vars_term_list' t) = vars_term t"
  by (induct t) simp_all
declare vars_term_list'.simps[simp del]

definition vars_term_list :: "('f, 'v) term \<Rightarrow> 'v list" where
  "vars_term_list = remdups \<circ> vars_term_list'"

lemma vars_term_list_sound[simp]:
  "set (vars_term_list t) = vars_term t"
  by (simp add: vars_term_list_def)

text {*The list of function symbols in a term (without removing duplicates).*}

fun funs_term_list' :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list' (Var _) = []"
| "funs_term_list' (Fun f ts) = f # concat (map funs_term_list' ts)"

lemma funs_term_list'_sound[simp]:
  "set (funs_term_list' t) = funs_term t"
  by (induct t) simp_all
declare funs_term_list'.simps[simp del]

definition funs_term_list :: "('f, 'v) term \<Rightarrow> 'f list" where
  "funs_term_list \<equiv> remdups \<circ> funs_term_list'"

lemma funs_term_list_sound[simp]:
  "set (funs_term_list t) = funs_term t"
  by (simp add: funs_term_list_def)


text {*The list of function symbol and arity pairs in a term
(without removing duplicates).*}

fun funas_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list' (Var _) = []"
| "funas_term_list' (Fun f ts) = (f, length ts) # concat (map funas_term_list' ts)"

lemma funas_term_list'_sound[simp]:
  "set (funas_term_list' t) = funas_term t"
  by (induct t) simp_all
declare funas_term_list'.simps[simp del]

definition funas_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list" where
  "funas_term_list \<equiv> remdups \<circ> funas_term_list'"

lemma funas_term_list_sound[simp]:
  "set (funas_term_list t) = funas_term t"
  by (simp add: funas_term_list_def)

lemma vars_term_list'_map_term: "vars_term_list' (map_term fg t) = vars_term_list' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case by (simp add: vars_term_list'.simps o_def, insert Fun, induct ss, auto)
qed

lemma funs_term_list'_map_term:
  "funs_term_list' (map_term fg t) = map fg (funs_term_list' t)"
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto)
qed

lemma linear_vars_term_list': assumes "linear_term t"
  shows "length (filter (op = x) (vars_term_list' t)) \<le> 1"
  using assms
proof (induct t)
  case (Var y)
  show ?case by (auto simp: vars_term_list'.simps)
next
  case (Fun f ss)
  show ?case 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from this[unfolded vars_term_list'.simps]
    have len: "2 \<le> length (filter (op = x) (concat (map vars_term_list' ss)))" (is "_ \<le> length ?xs") by auto
    from len obtain y ys where xs: "?xs = y # ys" by (cases ?xs, auto)
    from len[unfolded xs] obtain z zs where ys: "ys = z # zs" by (cases ys, auto)
    from xs[unfolded ys] have "{y,z} \<subseteq> set ?xs" by auto
    from this[unfolded set_filter] have "y = x" and "z = x" by auto
    from xs[unfolded ys this] have xs: "?xs = x # x # zs" by auto
    {
      fix s
      assume s: "s \<in> set ss"
      with Fun(2)[unfolded linear_term.simps list_all_iff]
      have "linear_term s" by auto
      note Fun(1)[OF s this]
    }
    from this Fun(2)[unfolded linear_term.simps list_all_iff] xs
    show False
    proof (induct ss)
      case Nil thus ?case by simp
    next
      case (Cons s ss) note oCons = this
      from Cons(3) have part: "is_partition (map vars_term ss) \<and> (\<forall> s \<in> set ss. linear_term s)" and lin: "linear_term s"  using is_partition_Cons by auto
      from Cons(1)[OF _ part] Cons(2) 
      have ind: "filter (op = x) (concat (map vars_term_list' ss)) \<noteq> x # x # zs" by auto
      show ?case 
      proof (cases "filter (op = x) (vars_term_list' s)")
        case Nil
        with Cons(4) ind show False by auto
      next
        case (Cons y ys)
        let ?s = "filter (op = x) (vars_term_list' s)"
        let ?ss = "filter (op = x) (concat (map vars_term_list' ss))"
        from Cons oCons(4) have "?s = x # ys" by auto
        with oCons(2)[of s] have sx: "?s = [x]"
          by auto
        with oCons(4) have ssx: "?ss = x # zs" by auto
        from sx have "x \<in> set ?s" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        have sx: "x \<in> vars_term s" by auto 
        from ssx have "x \<in> set ?ss" by auto
        from this[unfolded set_filter vars_term_list'_sound]
        obtain t where tx: "x \<in> vars_term t" and t: "t \<in> set ss" by auto
        from t[unfolded set_conv_nth] obtain i where i: "i < length ss" and
          t: "t = ss ! i" by auto
        from oCons(3)[THEN conjunct1, unfolded is_partition_def, THEN spec[of _ "Suc i"], THEN mp, THEN spec[of _ 0]] i tx[unfolded t] sx
        show False by auto
      qed
    qed  
  qed
qed


lemma vars_term_list_map_term:
  "vars_term_list (map_term fg t) = vars_term_list t"
  unfolding vars_term_list_def vars_term_list'_map_term o_def ..

fun supteq_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "supteq_list (Var x) = [Var x]"
| "supteq_list (Fun f ts) = Fun f ts # concat (map supteq_list ts)"


lemma supteq_list[simp]: "set (supteq_list t) = {s. t \<unrhd> s}"
proof (rule set_eqI, simp)
  fix s
  show "s \<in> set(supteq_list t) = (t \<unrhd> s)"
  proof (induct t, simp add: supteq_var_imp_eq)
    case (Fun f ss)
    show ?case
    proof (cases "Fun f ss = s", (auto)[1])
      case False
      show ?thesis
      proof
        assume "Fun f ss \<unrhd> s"
        with False have sup: "Fun f ss \<rhd> s" using supteq_supt_conv by auto
        obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" using sup by auto
        then obtain b D a where "Fun f ss = Fun f (b @ D\<langle>s\<rangle> # a)" by (cases C, auto)
        hence D: "D\<langle>s\<rangle> \<in> set ss" by auto
        with Fun[OF D] ctxt_imp_supteq[of D s] obtain t where "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        thus "s \<in> set (supteq_list (Fun f ss))" by auto
      next
        assume "s \<in> set (supteq_list (Fun f ss))"
        with False obtain t where t: "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        with Fun[OF t] have "t \<unrhd> s" by auto
        with t show "Fun f ss \<unrhd> s" by auto
      qed
    qed
  qed
qed

fun
  supt_impl :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "supt_impl (Var x) t = False" |
  "supt_impl (Fun f ss) t = (t \<in> set ss \<or> (list_ex (\<lambda>s. supt_impl s t) ss))"

lemma supt_impl[code]: "op \<rhd> = (\<lambda>(x, y). supt_impl x y)"
proof (intro equalityI subsetI2)
  fix s t assume "s \<rhd> t" hence "supt_impl s t"
  proof (induct s)
    case (Var x) thus ?case by auto
  next
    case (Fun f ss) thus ?case
    proof (cases "t \<in> set ss")
      case True thus ?thesis by (simp)
    next
      case False
      assume "\<And>s. \<lbrakk>s \<in> set ss; s \<rhd> t\<rbrakk> \<Longrightarrow> supt_impl s t"
        and "Fun f ss \<rhd> t" and "t \<notin> set ss"
      moreover from `Fun f ss \<rhd> t` obtain s where "s \<in> set ss" and "s \<rhd> t"
        by (cases rule: supt.cases) (simp_all add: `t \<notin> set ss`)
      ultimately have "supt_impl s t" by simp
      with `s \<in> set ss` show ?thesis by (auto simp: list_ex_iff)
    qed
  qed
  thus "(s, t) \<in> (\<lambda>(x, y). supt_impl x y)" by (simp add: mem_def split_def)
next
  fix s t assume "(s, t) \<in> (\<lambda>(x, y). supt_impl x y)"
  hence "supt_impl s t" by (simp add: mem_def)
  thus "s \<rhd> t"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss)
    hence "t \<in> set ss \<or> (\<exists>s\<in>set ss. supt_impl s t)" by (simp add: list_ex_iff)
    thus ?case
    proof
      assume "t \<in> set ss" thus ?case by (auto intro: supt.intros)
    next
      assume "\<exists>s\<in>set ss. supt_impl s t"
      then obtain s where "s \<in> set ss" and "supt_impl s t" by auto
      with Fun have "s \<rhd> t" by auto
      with `s \<in> set ss` show ?thesis by (auto intro: supt.intros)
    qed
  qed
qed
lemma supteq_impl[code]: "op \<unrhd> = (\<lambda>(x, y). x = y \<or> supt_impl x y)"
unfolding supteq_supt_set_conv by (auto simp: supt_impl mem_def)

fun
  linear_term_impl :: "'v list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v list) option"
where
  "linear_term_impl xs (Var x) = (if x \<in> set xs then None else Some (x # xs))" |
  "linear_term_impl xs (Fun _ []) = Some xs" |
  "linear_term_impl xs (Fun f (t # ts)) = (case linear_term_impl xs t of 
    None \<Rightarrow> None
  | Some ys \<Rightarrow> linear_term_impl ys (Fun f ts))"

lemma [code]: "linear_term t = (case linear_term_impl [] t of None \<Rightarrow> False | Some _ \<Rightarrow> True)" 
proof -
  {
    fix xs ys
    have "(linear_term_impl xs t = None \<longrightarrow> (set xs \<inter> vars_term t \<noteq> {} \<or> \<not> linear_term t)) \<and> 
      (linear_term_impl xs t = Some ys \<longrightarrow> (set ys = set xs \<union> vars_term t) \<and> set xs \<inter> vars_term t = {} \<and> linear_term t)"
    proof (induct arbitrary: ys rule: linear_term_impl.induct, force, simp add: is_partition_Nil)
      case (3 xs f t ts zs)
      show ?case 
      proof (cases "linear_term_impl xs t")
        case None
        with 3 show ?thesis by auto
      next
        case (Some ys)
        note some = this
        with 3 have rec1: "set ys = set xs \<union> vars_term t \<and> set xs \<inter> vars_term t = {} \<and> linear_term t" by auto
        show ?thesis 
        proof (cases "linear_term_impl ys (Fun f ts)")
          case None 
          with rec1 Some have res: "linear_term_impl xs (Fun f (t # ts)) = None" by simp
          from None 3(2) Some have rec2: "set ys \<inter> vars_term (Fun f ts) \<noteq> {} \<or> \<not> linear_term (Fun f ts)" by simp
          hence "set xs \<inter> vars_term (Fun f (t # ts)) \<noteq> {} \<or> \<not> linear_term (Fun f (t # ts))" 
          proof 
            assume "set ys \<inter> vars_term (Fun f ts) \<noteq> {}" 
            then obtain x where x1: "x \<in> set ys" and x2: "x \<in> vars_term (Fun f ts)" by auto
            show ?thesis 
            proof (cases "x \<in> set xs")
              case True
              with x2 show ?thesis by auto
            next
              case False
              with x1 rec1 have "x \<in> vars_term t" by auto
              with x2 have "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
              thus ?thesis ..
            qed
          next
            assume "\<not> linear_term (Fun f ts)" hence "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
            thus ?thesis ..
          qed            
          with res show ?thesis by auto
        next
          case (Some us)
          with some have res: "linear_term_impl xs (Fun f (t # ts)) = Some us" by auto          
          {
            assume us: "us = zs"
            from Some[simplified us] 3(2) some 
            have rec2: "set zs = set ys \<union> vars_term (Fun f ts) \<and> set ys \<inter> vars_term (Fun f ts) = {} \<and> linear_term (Fun f ts)" by auto
            from rec1 rec2  
            have part1: "set zs = set xs \<union> vars_term (Fun f (t # ts)) \<and> set xs \<inter> vars_term (Fun f (t # ts)) = {}" (is ?part1) by auto
            from rec1 rec2 have "vars_term t \<inter> vars_term (Fun f ts) = {}" and "linear_term t" and "linear_term (Fun f ts)" by auto
            hence "linear_term (Fun f (t # ts))" (is ?part2) by (auto simp: is_partition_Cons) 
            with part1 have "?part1 \<and> ?part2" ..
          }
          with res show ?thesis by auto
        qed
      qed
    qed
  } note main = this
  from main[of "[]"] show ?thesis by (cases "linear_term_impl [] t", auto)
qed

definition
  check_no_var :: "('f::show, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_var t \<equiv> check (\<not> is_var t) (''variable found'' +#+ shows_nl)"

lemma check_no_var_sound[simp]:
  "isOK (check_no_var t) = (\<not> is_var t)"
using assms unfolding check_no_var_def by simp


definition
  check_supt :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supt s t \<equiv> check (s \<rhd> t) (shows t +@+ '' is not a proper subterm of '' +#+ shows s)"

definition
  check_supteq :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supteq s t \<equiv> check (s \<unrhd> t) (shows t +@+ '' is not a subterm of '' +#+ shows s)"

lemma isOK_check_supt[simp]:
  "isOK (check_supt s t) = ((s, t) \<in> op \<rhd>)"
unfolding check_supt_def using supt_def by auto

lemma isOK_check_supteq[simp]:
  "isOK (check_supteq s t) = ((s, t) \<in> op \<unrhd>)"
unfolding check_supteq_def using supteq_def by auto

subsection {*Additional Functions on Terms*}

fun with_arity :: "('f, 'v) term \<Rightarrow> ('f \<times> nat, 'v) term" where
  "with_arity (Var x) = Var x"
| "with_arity (Fun f ts) = Fun (f, length ts) (map with_arity ts)"

fun add_vars :: "('f, 'v) term \<Rightarrow> 'v list \<Rightarrow> 'v list" where
  "add_vars (Var x) xs = x # xs"
| "add_vars (Fun _ ts) xs = foldr add_vars ts xs"

fun add_funs :: "('f, 'v) term \<Rightarrow> 'f list \<Rightarrow> 'f list" where
  "add_funs (Var _) fs = fs"
| "add_funs (Fun f ts) fs = f # foldr add_funs ts fs"

(*FIXME: define vars and funs via add_vars and add_funs*)
(*
abbreviation vars where "vars t \<equiv> add_vars t []"
abbreviation vars where "funs t \<equiv> add_funs t []"
*)
(*FIXME: define funas via with_arity and add_funs*)

lemma with_arity_term_fold[code]:
  "with_arity = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts)"
proof
  fix t show "with_arity t = Term.fold Var (\<lambda>f ts. Fun (f, length ts) ts) t"
    by (induct t) simp_all
qed

subsection {*Code Generator Setup*}

text{*The target languagte is assumed to be Haskell. Hence the following
code equations try to increase lazyness as far as possible.*}
lemma [code]:
  "add_vars = Term.fold (op #) (const (flip (foldr (op \<circ>)) id))"
    (is "_ = ?fun")
proof (intro ext)
  fix t xs show "add_vars t xs = ?fun t xs"
  proof (induct t arbitrary: xs)
    case (Var x) show ?case by simp
  next
    case (Fun f ts xs)
    from foldr_cong[of xs xs ts ts add_vars, OF _ _ this]
      have IH: "foldr add_vars ts xs = foldr ?fun ts xs" by blast
    show ?case
      by (simp add: IH foldr_map const_def del: o_apply id_apply)
  qed
qed

lemma [code]:
  "add_funs = Term.fold (const id) (\<lambda>f xs. op # f \<circ> foldr (op \<circ>) xs id)"
    (is "_ = ?fun")
proof (intro ext)
  fix t fs show "add_funs t fs = ?fun t fs"
  proof (induct t arbitrary: fs)
    case (Var x) show ?case by (simp add: const_def)
  next
    case (Fun f ts fs)
    from foldr_cong[of fs fs ts ts add_funs, OF _ _ this]
      have IH: "foldr add_funs ts fs = foldr ?fun ts fs" by blast
    show ?case
      by (simp add: IH const_def foldr_map del: o_apply id_apply)
         simp
  qed
qed

end
